import time

from d8s_fun import password_create, spin_until_done


def test_password_create_1():
    assert len(password_create()) == 15
    assert len(password_create(length=20)) == 20

    new_pwd = password_create(length=21, character_set='ab')
    assert len(new_pwd) == 21
    new_pwd_has_only_a_b = new_pwd.count('a') + new_pwd.count('b') == 21
    assert new_pwd_has_only_a_b


@spin_until_done
def spin_until_done_test_func_a(a):
    """."""
    time.sleep(1)
    return a


def test_spin_until_done_1():
    import timeit

    time_duration = timeit.timeit('spin_until_done_test_func_a(1)', number=3, globals=globals())
    assert 1 < time_duration > 2



import functools
import random
import time

PASSWORD_CHARACTER_SET = (
    '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!#$%&()*+,-.:;<=>?@[]^_`{|}~'  # nosec
)


def spinner(cycles: int = 10, *, frames_per_second: int = 10) -> None:
    """Print a spinner."""
    for i in range(0, cycles * 4):
        time.sleep(1 / frames_per_second)
        if i % 4 == 0:
            print('\r|', end='')
        elif i % 4 == 1:
            print('\r/', end='')
        elif i % 4 == 2:
            print('\r-', end='')
        else:
            print('\r\\', end='')
    print('\r|', end='')


def assumption_make():
    options = (
        'we have a can opener',
        'normal distribution',
        'that I exist',
        'air is diatomic',
        'that assumptions are dangerous',
        'coffee',
        'that all assumptions are invalid',
    )

    print('I\'ve assumed {}.'.format(random.choice(options)))  # nosec


def password_create(*, length: int = 15, character_set: str = PASSWORD_CHARACTER_SET) -> str:
    """Create a password of the given length using the given character_set."""
    return ''.join(random.choices(character_set, k=length))


def spin_until_done(func):
    """Show a spinner until the function is done."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        import concurrent.futures

        with concurrent.futures.ThreadPoolExecutor() as executor:
            executor.submit(spinner)
            function_submission = executor.submit(func, *args, **kwargs)
            # TODO: add some stop here to make sure that the spinner does not just keep running
            while not function_submission.done():
                spinner()
            return function_submission.result()

    return wrapper



from collections import deque
from typing import Any

import pytest
from d8s_file_system import file_read, file_write
from d8s_python import python_function_blocks
from d8s_strings import string_remove_after, string_remove_from_start

FILL_OUTPUT_SIGNAL = 'fill'
FILL_OUTPUT_SIGNAL_IN_CONTEXT = f'== {FILL_OUTPUT_SIGNAL.__repr__()}'
RAISES_OUTPUT_SIGNAL = 'raises'
RAISES_OUTPUT_SIGNAL_IN_CONTEXT = f'== {RAISES_OUTPUT_SIGNAL.__repr__()}'

fill_values = deque()


def pytest_assertrepr_compare(config, op: str, left: object, right: object):
    global fill_values

    if right == FILL_OUTPUT_SIGNAL:
        fill_values.append(left)
        pytest.skip(f'{op} {right.__repr__()}')


def _update_test(file_path: str, function_name: str, new_assertion_value: Any) -> bool:
    """Update the test with the given function_name in the file at the given file_path with the new_assertion_value."""
    test_file_content = file_read(file_path)
    original_function_block = function_block = [
        i for i in python_function_blocks(test_file_content) if function_name in i
    ][0]
    function_block = function_block.replace(FILL_OUTPUT_SIGNAL_IN_CONTEXT, f'== {new_assertion_value.__repr__()}', 1)
    test_file_content = test_file_content.replace(original_function_block, function_block, 1)
    return file_write(file_path, test_file_content)


def _update_test_with_error(file_path: str, function_name: str, error_type: str, erroneous_assertion: str) -> bool:
    """Update the test with the given function_name in the file at the given file_path with the new_assertion_value."""
    test_file_content = file_read(file_path)
    original_function_block = function_block = [
        i for i in python_function_blocks(test_file_content) if function_name in i
    ][0]
    # TODO: currently, the indentation of the new assertion is hard coded; eventually, I would like to get the indentation from the original assertion
    new_assertion = f'with pytest.raises({error_type}):\n        {erroneous_assertion}'
    full_erroneous_assertion = f'assert {erroneous_assertion} == \'{RAISES_OUTPUT_SIGNAL}\''
    function_block = function_block.replace(full_erroneous_assertion, new_assertion, 1)
    test_file_content = test_file_content.replace(original_function_block, function_block, 1)
    return file_write(file_path, test_file_content)


def _get_erroneous_call(report_text: str) -> str:
    """."""
    erroneous_line = [
        line for line in report_text.splitlines() if line.startswith('> ') and RAISES_OUTPUT_SIGNAL_IN_CONTEXT in line
    ][0]
    erroneous_assertion = erroneous_line.lstrip('> ')
    erroneous_assertion = string_remove_from_start(erroneous_assertion, 'assert ')
    erroneous_assertion = string_remove_after(erroneous_assertion, ' ==')
    erroneous_assertion = erroneous_assertion.rstrip('= ')
    return erroneous_assertion


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    global fill_values

    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # we only look at actual failing test calls, not setup/teardown
    if rep.when == "call" and rep.skipped:
        if fill_values and (FILL_OUTPUT_SIGNAL_IN_CONTEXT in rep.longreprtext):
            file_path = rep.fspath
            function_name = item.name
            new_assertion_value = fill_values.popleft()
            _update_test(file_path, function_name, new_assertion_value)
    elif rep.when == 'call' and rep.failed:
        if RAISES_OUTPUT_SIGNAL_IN_CONTEXT in rep.longreprtext:
            print('here!!!')
            file_path = rep.fspath
            function_name = item.name
            error_type = rep.longreprtext.split(" ")[-1]
            erroneous_assertion = _get_erroneous_call(rep.longreprtext)
            _update_test_with_error(file_path, function_name, error_type, erroneous_assertion)



#!/usr/bin/env python

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('requirements.txt') as requirements_file:
    requirements = requirements_file.read().splitlines()

with open('requirements_dev.txt') as requirements_dev_file:
    test_requirements = requirements_dev_file.read().splitlines()

setup(
    name='d8s_fun',
    description="Fun Democritus functions.",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Floyd Hightower",
    author_email='floyd.hightower27@gmail.com',
    url='https://github.com/democritus-project/d8s-fun',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    packages=find_packages(exclude=('tests')),
    include_package_data=True,
    install_requires=requirements,
    license="GNU Lesser General Public License v3",
    zip_safe=True,
    keywords="democritus,utility,python,fun-things,fun-things-utility,fun,joy,easter-eggs",
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    test_suite='tests',
    tests_require=test_requirements,
)
